/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * ConfigLoaderXML.java
 *
 * Created on 19 July 2008, 05:22
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package net.infantai.dbcp.secure;


import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class ConfigLoaderXML extends DefaultHandler implements ConfigLoader
{
    
//    private static final Logger logger = LogManager.getLogger(ConfigLoaderXML.class);
    
    private String dbDriverClassName = "org.postgresql.Driver";
    private String poolIdentifier = "";
    private String dbServer = "";
    private String dbServerPort = "";
    private String dbName = "";
    private String dbUsername;
    private String dbPassword;
    private String dbConnectionString = "";    
    
    // Database Connection Pool Parameters
    private int dbConnPoolMaxPoolSize = 20;
    private int dbConnPoolMaxWait = 8000; // Milliseconds        
    
    private String configXMLFileName;
    private String tagValue = "";
    private DBServerConfig dbserverconfig = null;
    private List<DBServerConfig> dbservers = null;
    
    private boolean isDBSettingsConfig = false;
    private boolean isDBDriverClassName = false;
    private boolean isDBServer = false;
    private boolean isDBServerPort = false;
    private boolean isDBName = false;
    private boolean isDBUsername = false;
    private boolean isDBPassword = false;
    private boolean isDBConnectionString = false;    
    private boolean isDBConnPoolMaxPoolSize = false;
    private boolean isDBConnPoolMaxWait = false;    
    private boolean isPoolIdentifier = false;
    
    public ConfigLoaderXML() 
    {
        configXMLFileName = "dbcpconfig.xml";
        dbservers = new ArrayList<DBServerConfig>();
        loadConfiguration();
    }
    
    private void loadConfiguration()
    {
        SAXParserFactory spf = SAXParserFactory.newInstance();
        try
        {
            SAXParser sp = spf.newSAXParser();
            sp.parse(configXMLFileName, this);
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
    }
    
    @Override
    public void startElement (String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        tagValue = "";
        if (qName.equals("dbsettings"))
        {
            dbserverconfig = new DBServerConfig();
            isDBSettingsConfig = true;
        }
        else if (qName.equals("poolidentifier") && isDBSettingsConfig)
        {
            isPoolIdentifier = true;
        }        
        else if (qName.equals("dbdriverclass") && isDBSettingsConfig)
        {
            isDBDriverClassName = true;
        }
        else if (qName.equals("dbserver") && isDBSettingsConfig)
        {
            isDBServer = true;
        }
        else if (qName.equals("dbserverport") && isDBSettingsConfig)
        {
            isDBServerPort = true;
        }
        else if (qName.equals("dbname") && isDBSettingsConfig)
        {
            isDBName = true;
        }
        else if (qName.equals("dbusername") && isDBSettingsConfig)
        {
            isDBUsername = true;
        }
        else if (qName.equals("dbpassword") && isDBSettingsConfig)
        {
            isDBPassword = true;
        }
        else if (qName.equals("dbconnectionstring") && isDBSettingsConfig)
        {
            isDBConnectionString = true;
        }        
        else if (qName.equals("maxpoolsize") && isDBSettingsConfig)
        {
            isDBConnPoolMaxPoolSize = true;
        }
        else if (qName.equals("maxwait") && isDBSettingsConfig)
        {
            isDBConnPoolMaxWait = true;
        }        
    }
    
    @Override
    public void endElement (String uri, String localName, String qName) throws SAXException
    {
        if (qName.equals("dbsettings"))
        {
            dbservers.add(this.dbserverconfig);
            isDBSettingsConfig = false;
        }
        else if (qName.equals("poolidentifier") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                poolIdentifier = tagValue;
                if (this.dbserverconfig != null)
                {
                    this.dbserverconfig.setPoolIdentifier(poolIdentifier);
                }
            }            
            isDBDriverClassName = false;
        }        
        else if (qName.equals("dbdriverclass") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbDriverClassName = tagValue;
                if (this.dbserverconfig != null)
                {
                    this.dbserverconfig.setDbDriverClassName(dbDriverClassName);
                }
            }            
            isDBDriverClassName = false;
        }
        else if (qName.equals("dbserver") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbServer = tagValue;
                if (this.dbserverconfig != null)
                {
                    this.dbserverconfig.setDbServer(dbServer);
                }                
            }            
            isDBServer = false;
        }
        else if (qName.equals("dbserverport") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbServerPort = tagValue;
                if (this.dbserverconfig != null)
                {
                    this.dbserverconfig.setDbServerPort(dbServerPort);
                } 
            }            
            isDBServerPort = false;
        }
        else if (qName.equals("dbname") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbName = tagValue;
                if (this.dbserverconfig != null)
                {
                    this.dbserverconfig.setDbInstance(dbName);
                }                
            }            
            isDBName = false;
        }
        else if (qName.equals("dbusername") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbUsername = tagValue;
                dbserverconfig.setDbUsername(dbUsername);
            }            
            isDBUsername = false;
        }
        else if (qName.equals("dbpassword") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbPassword = tagValue;
                dbserverconfig.setDbPassword(dbPassword);
            }
            isDBPassword = false;
        }
        else if (qName.equals("dbconnectionstring") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbConnectionString = tagValue;
                dbserverconfig.setDbConnectionString(dbConnectionString);
            }            
            isDBConnectionString = false;
        }
        else if (qName.equals("maxpoolsize") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbConnPoolMaxPoolSize = Integer.parseInt(tagValue);
            }            
            isDBConnPoolMaxPoolSize = false;
        }
        else if (qName.equals("maxwait") && isDBSettingsConfig)
        {
            if (tagValue.length() > 0)
            {
                dbConnPoolMaxWait = Integer.parseInt(tagValue);
            }            
            isDBConnPoolMaxWait = false;
        }
    }
    
    @Override
    public void characters (char ch[], int start, int length) throws SAXException
    {
        tagValue = tagValue + new String(ch, start, length);
    }

    public List<DBServerConfig> getDBServerConfig()
    {
        return dbservers;
    }
    
}
