/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.dbcp.secure;

import net.infantai.util.common.FileIO;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.GregorianCalendar;
import java.util.List;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class DBConnSecure 
{
    
    // Very basic class to establish connectivity with a database.
    
    private static DBConnSecure dbconn = null;
    
    private static Connection connection = null;
    
    private DBConnSecure() throws ClassNotFoundException, java.sql.SQLException, IOException
    {
        String filename = "dbconparamsecure";
        
        String driverClassName = null;
        String dburl = null;
        
        // read file dbconparam from classpath
        List<String> dbParamList = FileIO.readTextFileAsList(filename);
        if (dbParamList == null)
        {
            throw new IOException("Error reading file " + filename + ": Please make sure this file is available on the classpath !");
        }
        
        for (int i = 0; i < dbParamList.size(); ++i)
        {
            String strLine = dbParamList.get(i);
            String[] nameValue = strLine.split("=", 2);
            if (nameValue.length == 2)
            {
                if (nameValue[0].equalsIgnoreCase("driver"))
                {
                    driverClassName = nameValue[1].trim();
                }
                else if (nameValue[0].equalsIgnoreCase("url"))
                {
                    dburl = nameValue[1].trim();
                }
            }
        }
        
        if (driverClassName == null || dburl == null)
        {
            throw new SQLException("Insufficient parameters in " + filename + " !");
        }
        
        //
        
        Class.forName(driverClassName);
        connection = DriverManager.getConnection(dburl);
    }
    
    public static DBConnSecure getInstance() throws ClassNotFoundException, java.sql.SQLException, IOException
    {
        if (dbconn == null)
        {
            dbconn = new DBConnSecure();
        }
        return dbconn;
    }
    
    /*
    public Connection getConnection()
    {
        return connection;
    }
     * 
     */
    
    public Connection getConnection()
    {
        try
        {
            if (connection == null || connection.isClosed())
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Warning: DB Connection closed, re-initializing singleton object.");
                dbconn = new DBConnSecure(); // re-instantiate the singleton object of this class: possibly unsafe call
            }        
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return connection;
    }    
    
}
