/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import net.infantai.util.common.FileIO;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class IWLLoader
{
    
    private static final Logger logger = LogManager.getLogger(OrganizationNameTagger.class);
    
    private static IWLLoader iwlLoader = null;
    private static final Object singletonLock = new Object();
    
    private static Set<String> prepositions = new HashSet<>();
    private static Set<String> conjunctions = new HashSet<>();
    private static Set<String> pronouns = new HashSet<>();
    private static Set<String> interjections = new HashSet<>();
    private static Set<String> articles = new HashSet<>();
    private static Set<String> firstNames = new HashSet<>();
    private static Set<String> lastNames = new HashSet<>();
    
    private IWLLoader()
    {
        try
        {
            loadInfantWordLists();
        }
        catch (Exception ex)
        {
            logger.error(ex.getMessage(), ex);
        }
    }
    
    public static IWLLoader getInstance()
    {
        if (iwlLoader == null)
        {
            iwlLoader = new IWLLoader();
        }
        return iwlLoader;
    }
    
    private static void loadInfantWordLists() throws IOException
    {
        prepositions = FileIO.readTextFileAsSet("models/infant/iwl/prepositions.iwl");
        conjunctions = FileIO.readTextFileAsSet("models/infant/iwl/conjunctions.iwl");
        pronouns = FileIO.readTextFileAsSet("models/infant/iwl/pronouns.iwl");
        interjections = FileIO.readTextFileAsSet("models/infant/iwl/interjections.iwl");
        articles = FileIO.readTextFileAsSet("models/infant/iwl/articles.iwl");
//        firstNames = FileIO.readTextFileAsSet("models/infant/iwl/firstnames.iwl");
//        lastNames = FileIO.readTextFileAsSet("models/infant/iwl/lastnames.iwl");
    }
    
    /**
     * @return the prepositions
     */
    public Set<String> getPrepositions()
    {
        return prepositions;
    }

    /**
     * @return the conjunctions
     */
    public Set<String> getConjunctions()
    {
        return conjunctions;
    }
    
    /**
     * @return the pronouns
     */
    public Set<String> getPronouns()
    {
        return pronouns;
    }

    /**
     * @return the interjections
     */
    public Set<String> getInterjections()
    {
        return interjections;
    }

    /**
     * @return the articles
     */
    public Set<String> getArticles()
    {
        return articles;
    }    

    /**
     * @return the firstNames
     */
    public Set<String> getFirstNames()
    {
        return firstNames;
    }

    /**
     * @return the lastNames
     */
    public Set<String> getLastNames()
    {
        return lastNames;
    }    
    
}
