/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp;

import java.util.Iterator;
import java.util.Set;
import net.infantai.nlp.ner.EntityTaggerOutput;
import net.infantai.util.common.TextUtil;
import net.infantai.util.common.TokenGroup;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class NegationDetector
{
    
    private static final String STR_REGEX_FOR_NEGATION_DELIMITERS = "[^a-zA-Z0-9'’]+";
    private static final TextUtil tu = new TextUtil(STR_REGEX_FOR_NEGATION_DELIMITERS); // do not split on apostrophes
    
    public static TokenGroup getNegationText(String input, EntityTaggerOutput eto, Set<String> negationPreMarkers, int maxTokensToPre, Set<String> negationPostMarkers, int maxTokensPost)
    {
        String negationText = null;
        TokenGroup tgNegation = null;
        int negationMarkerIndex = -1;
        String left = input.substring(0, eto.tg.getIndex());
        Set<String> preTokens = tu.getUpToNTokensFromPrevText(left, maxTokensToPre);
        Iterator<String> iterPreTokens = preTokens.iterator();
        while (iterPreTokens.hasNext())
        {
            String preText = iterPreTokens.next();
            if (negationPreMarkers.contains(preText))
            {
                String[] matchedParts = preText.split(STR_REGEX_FOR_NEGATION_DELIMITERS);
                negationMarkerIndex = left.lastIndexOf(matchedParts[0]);
                negationText = preText;
                break;
            }
        }
        
        if (negationText == null) // pre-negation not found, look for post
        {
            int startIndex = eto.tg.getIndex() + eto.tg.getSize() + 1;
            if (startIndex < input.length())
            {
                String right = input.substring(startIndex);
                Set<String> postTokens = tu.getUpToNTokensFromNextText(right, maxTokensPost);
                Iterator<String> iterPostTokens = postTokens.iterator();
                while (iterPostTokens.hasNext())
                {
                    String postText = iterPostTokens.next();
                    if (negationPostMarkers.contains(postText))
                    {
                        String[] matchedParts = postText.split(STR_REGEX_FOR_NEGATION_DELIMITERS);
                        negationMarkerIndex = startIndex + right.indexOf(matchedParts[0]);                        
                        negationText = postText;
                        break;
                    }
                }            
            }
        }
        
        if (negationText != null)
        {
            tgNegation = new TokenGroup(negationText, negationMarkerIndex, negationText.length());
        }
        
        return tgNegation;
    }
    
}
