/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.GregorianCalendar;
import opennlp.tools.sentdetect.SentenceDetectorME;
import opennlp.tools.sentdetect.SentenceModel;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class SentenceDetector
{
    private static final Logger logger = LogManager.getLogger(SentenceDetector.class);
    private static SentenceModel sentenceModel = null;
    
    public SentenceDetector()
    {
        InputStream modelInStream = null;
        try
        {        
            modelInStream = new FileInputStream("./models/opennlp/en-sent.bin"); // from http://opennlp.sourceforge.net/models-1.5/
            sentenceModel = new SentenceModel(modelInStream);
            modelInStream.close();
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
            fnfe.printStackTrace();
            logger.error(fnfe.getMessage(), fnfe);
        }
        catch (IOException ioex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
            ioex.printStackTrace();
            logger.error(ioex.getMessage(), ioex);
        }
        finally
        {
            try
            {
                if (modelInStream != null)
                {
                    modelInStream.close();
                }
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                logger.error(ioex.getMessage(), ioex);                
            }
        }
    }
    
    public String[] detectSentences(String paragraph) throws IOException
    {
        opennlp.tools.sentdetect.SentenceDetector sentenceDetector = new SentenceDetectorME(sentenceModel);
        String sentences[] = sentenceDetector.sentDetect(paragraph);
        return sentences;
    }    
    
}
