/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp.categorization;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import net.infantai.util.common.BooleanTextRuleMatcher;
import net.infantai.util.common.MalformedExpressionException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class TextCategorization
{

    private static final Logger LOGGER = LogManager.getLogger(TextCategorization.class);

    private Map<Integer, Integer> textCategorizationKeywordIdCategoryId = new LinkedHashMap<Integer, Integer>();
    private Map<Integer, BooleanTextRuleMatcher> textCategorizationKeywordIdRuleBTRM = new LinkedHashMap<Integer, BooleanTextRuleMatcher>();

//    StaticLoader sl;

    public TextCategorization(Map<Integer, Integer> textCategorizationKeywordIdCategoryId, Map<Integer, BooleanTextRuleMatcher> textCategorizationKeywordIdRuleBTRM)
    {
//        sl = StaticLoader.getInstance();
        this.textCategorizationKeywordIdCategoryId = textCategorizationKeywordIdCategoryId;
        this.textCategorizationKeywordIdRuleBTRM = textCategorizationKeywordIdRuleBTRM;
    }

    public Set<TextCategorizationOutput> categorize(String input, Long inputId)
    {
        Set<TextCategorizationOutput> tagCloudOutputSet = new HashSet<>();
        try
        {
            LOGGER.debug("Running TagCloud category module...");
//            textCategorizationKeywordIdCategoryId = sl.getTextCategorizationKeywordIdCategoryId();
//            textCategorizationKeywordIdRuleBTRM = sl.getTextCategorizationKeywordIdRuleBTRM();

            // tag cloud keywords
            Set<String> articleTokens = BooleanTextRuleMatcher.getDistinctTokenCombinations(input.toLowerCase());
            Iterator<Map.Entry<Integer, BooleanTextRuleMatcher>> iterTagCloudRuleBTRM = textCategorizationKeywordIdRuleBTRM.entrySet().iterator();
            while (iterTagCloudRuleBTRM.hasNext())
            {
                Map.Entry<Integer, BooleanTextRuleMatcher> mapEntryBTRM = iterTagCloudRuleBTRM.next();
                Integer ruleId = mapEntryBTRM.getKey();
                BooleanTextRuleMatcher btrmRule = mapEntryBTRM.getValue();

                try
                {
                    if (btrmRule.matchBooleanTextRule(articleTokens))
                    {
                        Integer kwcatid = textCategorizationKeywordIdCategoryId.get(ruleId);
                        TextCategorizationOutput tcOutput = new TextCategorizationOutput();
//                        tcOutput.setInputId(apiRequest.getInputId());
                        tcOutput.setCategoryId(kwcatid);
                        tcOutput.setCategoryKeywordId(ruleId);
                        tagCloudOutputSet.add(tcOutput);
                    }                
                }
                catch (MalformedExpressionException meex)
                {
                    LOGGER.error("Error processing Rule: " + btrmRule.toString() + " {rule_id: " + ruleId + "}", meex);
                }
            }
            LOGGER.debug("Tag Cloud Output for the News Is:" );
        }
        catch (Exception ex)
        {
            LOGGER.error("Failed to run tagcloud on input: id = " + inputId, ex);
        }

        return tagCloudOutputSet;
    }
}
