/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp.ner;

//import com.newsanalysis.tagging.*;
import net.infantai.util.common.BooleanTextRuleMatcher;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class EntityKeyword
{

    private Long keywordId = -1L;
    private String keyword = null;
    private Long entityTypeId = -1L;
    private String entityTypeName = null;    
    private Long entityId = -1L;
    private String entityName = null;
    private BooleanTextRuleMatcher keywordRule = null;
    private boolean caseSensitive = false;
    private boolean isModification = false;
    private String keywordOriginal = null;
    private boolean attemptNumericValueExtraction = false;
    
    public EntityKeyword()
    {
    
    }
    
    public EntityKeyword(EntityKeyword ek)
    {
        if (ek != null)
        {
            this.caseSensitive = ek.isCaseSensitive();
            this.entityId = ek.getEntityId();
            this.entityName = ek.getEntityName();
            this.entityTypeId = ek.getEntityTypeId();
            this.entityTypeName = ek.getEntityTypeName();
            this.keyword = ek.getKeyword();
            this.keywordId = ek.getKeywordId();
            this.keywordRule = ek.getKeywordRule();
            this.caseSensitive = ek.caseSensitive;
            this.isModification = ek.isModification;
            this.keywordOriginal = ek.keywordOriginal;
            this.attemptNumericValueExtraction = ek.attemptNumericValueExtraction;
        }
    }

    /**
     * @return the keywordId
     */
    public Long getKeywordId()
    {
        return keywordId;
    }

    /**
     * @param keywordId the keywordId to set
     */
    public void setKeywordId(Long keywordId)
    {
        this.keywordId = keywordId;
    }
    
    /**
     * @return the keyword
     */
    public String getKeyword()
    {
        return keyword;
    }

    /**
     * @param keyword the keyword to set
     */
    public void setKeyword(String keyword)
    {
        this.keyword = keyword;
    }    

    /**
     * @return the entityTypeId
     */
    public Long getEntityTypeId()
    {
        return entityTypeId;
    }

    /**
     * @param entityTypeId the entityTypeId to set
     */
    public void setEntityTypeId(Long entityTypeId)
    {
        this.entityTypeId = entityTypeId;
    }

    /**
     * @return the entityTypeName
     */
    public String getEntityTypeName()
    {
        return entityTypeName;
    }

    /**
     * @param entityTypeName the entityTypeName to set
     */
    public void setEntityTypeName(String entityTypeName)
    {
        this.entityTypeName = entityTypeName;
    }
    
    /**
     * @return the entityId
     */
    public Long getEntityId()
    {
        return entityId;
    }

    /**
     * @param entityId the entityId to set
     */
    public void setEntityId(Long entityId)
    {
        this.entityId = entityId;
    }    
    
    /**
     * @return the entityName
     */
    public String getEntityName()
    {
        return entityName;
    }

    /**
     * @param entityName the entityName to set
     */
    public void setEntityName(String entityName)
    {
        this.entityName = entityName;
    }    

    public BooleanTextRuleMatcher getKeywordRule()
    {
        return keywordRule;
    }

    public void setKeywordRule(BooleanTextRuleMatcher keywordRule)
    {
        this.keywordRule = keywordRule;
    }
    
    /**
     * @return the caseSensitive
     */
    public boolean isCaseSensitive()
    {
        return caseSensitive;
    }

    /**
     * @param caseSensitive the caseSensitive to set
     */
    public void setCaseSensitive(boolean caseSensitive)
    {
        this.caseSensitive = caseSensitive;
    }    
    
    /**
     * @return the isModification
     */
    public boolean getIsModification()
    {
        return isModification;
    }

    /**
     * @param isModification the isModification to set
     */
    public void setIsModification(boolean isModification)
    {
        this.isModification = isModification;
    }    
    
    /**
     * @return the keywordOriginal
     */
    public String getKeywordOriginal()
    {
        return keywordOriginal;
    }

    /**
     * @param keywordOriginal the keywordOriginal to set
     */
    public void setKeywordOriginal(String keywordOriginal)
    {
        this.keywordOriginal = keywordOriginal;
    }    
    
    /**
     * @return the attemptNumericValueExtraction
     */
    public boolean attemptNumericValueExtraction()
    {
        return attemptNumericValueExtraction;
    }

    /**
     * @param attemptNumericValueExtraction the attemptNumericValueExtraction to set
     */
    public void setAttemptNumericValueExtraction(boolean attemptNumericValueExtraction)
    {
        this.attemptNumericValueExtraction = attemptNumericValueExtraction;
    }    

    @Override
    public boolean equals(Object obj)
    {
        if (obj != null && obj instanceof EntityKeyword)
        {
            EntityKeyword ek = (EntityKeyword) obj;
            if (ek.getKeywordId().longValue() == this.getKeywordId().longValue() && ek.keyword.equals(this.keyword))
            {
                return true;
            }        
        }
        return false;
    }
    
    @Override
    public int hashCode()
    {
        return this.getKeywordId().intValue();
    }
    
}
