/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp.ner;

import java.util.Collections;
import java.util.List;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class OverlapResolver
{
    
    private static final Logger LOGGER = LogManager.getLogger(OverlapResolver.class);
    
    public static void resolveOverlaps(String input, List<? extends EntityTaggerOutput> allTaggedEntities)
    {
        Collections.sort(allTaggedEntities);
        if (allTaggedEntities.size() > 1)
        {
            for (int i = 0; i < (allTaggedEntities.size() - 1); ++i)
            {
                EntityTaggerOutput etoCurrent = allTaggedEntities.get(i);
                EntityTaggerOutput etoNext = allTaggedEntities.get(i + 1);
                if (etoCurrent.tg != null & etoNext.tg != null)
                {
                    if (etoCurrent.tg.getIndex().intValue() == etoNext.tg.getIndex().intValue())
                    {
                        if (etoCurrent.tg.getSize() > etoNext.tg.getSize())
                        {
                            EntityTaggerOutput etoToDiscard = allTaggedEntities.get(i + 1);
                            LOGGER.debug("Entity " + etoToDiscard.tg.getToken() + " at index " + etoToDiscard.tg.getIndex() + " (next entity) fails Overlap Rule, discarding.");                                                                                                        
                            allTaggedEntities.remove(i + 1);
                        }
                        else if (etoCurrent.tg.getSize() < etoNext.tg.getSize())
                        {
                            EntityTaggerOutput etoToDiscard = allTaggedEntities.get(i);
                            LOGGER.debug("Entity " + etoToDiscard.tg.getToken() + " at index " + etoToDiscard.tg.getIndex() + " (current entity) fails Overlap Rule, discarding.");                                                                            
                            allTaggedEntities.remove(i);
                            --i;
                        }
                    }
                    else if (etoNext.tg.getIndex() > etoCurrent.tg.getIndex())
                    {
                        if (etoNext.tg.getIndex() < (etoCurrent.tg.getIndex() + etoCurrent.tg.getSize()))
                        {
                            if (etoCurrent.ek != null && etoNext.ek != null)
                            {
                                if (etoCurrent.ek.getIsModification() == true && etoNext.ek.getIsModification() == false)
                                {
                                    // in case of overlaps, pick unmodified keywords over modified ones
                                    EntityTaggerOutput etoToDiscard = allTaggedEntities.get(i);
                                    LOGGER.debug("Entity " + etoToDiscard.tg.getToken() + " at index " + etoToDiscard.tg.getIndex() + " (current entity) fails Overlap Rule (prefer unmodified keyword over modified one), discarding.");                                                                            
                                    allTaggedEntities.remove(i);
                                    --i;
                                    continue;                                    
                                }
                            }
                            EntityTaggerOutput etoToDiscard = allTaggedEntities.get(i + 1);
                            LOGGER.debug("Entity " + etoToDiscard.tg.getToken() + " at index " + etoToDiscard.tg.getIndex() + " (next entity starts within current entity) fails Overlap Rule, discarding.");                                                                                                        
                            allTaggedEntities.remove(i + 1);
                        }
                    }
                }
            }
        }
    }
    
}
