/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.util.common;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class ByteInterpreter
{
    
    /**
     * Big-Endian Byte Ordering - Java / Network Byte Order (Left To Right / MSByte First / Lowest Memory Address First)
     */
    public static final int BYTE_ORDER_BIG_ENDIAN = 1;
    
    /**
     * Little-Endian Byte Ordering - Reverse of Big-Endian.
     */
    public static final int BYTE_ORDER_LITTLE_ENDIAN = 2;
    
    public static final String MESSAGE_ILLEGAL_OFFSET = "Illegal Offset!";
    public static final String MESSAGE_UNRECOGNIZED_BYTE_ORDER = "Unrecognized Byte Order!";
    
    public static void showBits(int value) // ha ha
    {
        System.out.print(Integer.toBinaryString(value));
    }
    
    public static int byteToUnsignedChar(byte b)
    {        
        int value = (int) (b & 0xFF); // Logical AND with bit-pattern 11111111 (as int) to widen range and retain the sign bit as part of the value. Shame on Java for making 'byte' a signed data type.
        return value;
    }    
    
    public static int byteToUnsignedChar(byte[] in, int position) throws ByteInterpretationException
    {
        if (in.length < position + 1)
        {
            throw new ByteInterpretationException(MESSAGE_ILLEGAL_OFFSET);
        }
        
        int valueShort = (int) (in[position] & 0xFF); // Logical AND with bit-pattern 11111111 (as int) to widen range and retain the sign bit as part of the value. Shame on Java for making 'byte' a signed data type.
        return valueShort;
    }
    
    public static short bytesToShort(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        if (in.length < offset + 2)
        {
            throw new ByteInterpretationException(MESSAGE_ILLEGAL_OFFSET);
        }
        
        short valueShort = 0;
        
        if (byteOrder == BYTE_ORDER_BIG_ENDIAN)
        {
            valueShort = (short) ((in[offset] & 0xFF) << 8 | (in[offset + 1] & 0xFF));
        }
        else if (byteOrder == BYTE_ORDER_LITTLE_ENDIAN)
        {
            valueShort = (short) ((in[offset + 1] & 0xFF) << 8 | (in[offset] & 0xFF));
        }
        else
        {
            throw new ByteInterpretationException(MESSAGE_UNRECOGNIZED_BYTE_ORDER);
        }
        
        return valueShort;
    }
    
    public static int bytesToUnsignedShort(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        if (in.length < offset + 2)
        {
            throw new ByteInterpretationException(MESSAGE_ILLEGAL_OFFSET);
        }
        
        int valueUShort = 0;
        
        if (byteOrder == BYTE_ORDER_BIG_ENDIAN)
        {
            valueUShort = (int) ((in[offset] & 0xFF) << 8 | (in[offset + 1] & 0xFF));
        }
        else if (byteOrder == BYTE_ORDER_LITTLE_ENDIAN)
        {
            valueUShort = (int) ((in[offset + 1] & 0xFF) << 8 | (in[offset] & 0xFF));
        }
        else
        {
            throw new ByteInterpretationException(MESSAGE_UNRECOGNIZED_BYTE_ORDER);
        }
        
        return valueUShort;
    }
    
    public static int bytesToInt(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        if (in.length < offset + 4)
        {
            throw new ByteInterpretationException(MESSAGE_ILLEGAL_OFFSET);
        }
        
        int valueInt = 0;
        
        if (byteOrder == BYTE_ORDER_BIG_ENDIAN)
        {
            valueInt = (int) ((in[offset] & 0xFF) << 24 | (in[offset + 1] & 0xFF) << 16 | (in[offset + 2] & 0xFF) << 8 | (in[offset + 3] & 0xFF));
        }
        else if (byteOrder == BYTE_ORDER_LITTLE_ENDIAN)
        {
            valueInt = (int) ((in[offset + 3] & 0xFF) << 24 | (in[offset + 2] & 0xFF) << 16 | (in[offset + 1] & 0xFF) << 8 | (in[offset] & 0xFF));
        }
        else
        {
            throw new ByteInterpretationException(MESSAGE_UNRECOGNIZED_BYTE_ORDER);
        }
        
        return valueInt;
    }
    
    public static long bytesToUnsignedInt(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        if (in.length < offset + 4)
        {
            throw new ByteInterpretationException(MESSAGE_ILLEGAL_OFFSET);
        }
        
        long valueUInt = 0;
        
        if (byteOrder == BYTE_ORDER_BIG_ENDIAN)
        {
            valueUInt = (long) (((long) in[offset] & 0xFF) << 24 | (in[offset + 1] & 0xFF) << 16 | (in[offset + 2] & 0xFF) << 8 | (in[offset + 3] & 0xFF));
        }
        else if (byteOrder == BYTE_ORDER_LITTLE_ENDIAN)
        {
            valueUInt = (long) (((long) in[offset + 3] & 0xFF) << 24 | (in[offset + 2] & 0xFF) << 16 | (in[offset + 1] & 0xFF) << 8 | (in[offset] & 0xFF));
        }
        else
        {
            throw new ByteInterpretationException(MESSAGE_UNRECOGNIZED_BYTE_ORDER);
        }
        
        return valueUInt;
    }    
    
    public static long bytesToLong(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        if (in.length < offset + 8)
        {
            throw new ByteInterpretationException(MESSAGE_ILLEGAL_OFFSET);
        }
        
        long valueLong = 0;
        
        if (byteOrder == BYTE_ORDER_BIG_ENDIAN)
        {
            valueLong = (long) (((long) in[offset] & 0xFF) << 56 | ((long) in[offset + 1] & 0xFF) << 48 | ((long) in[offset + 2] & 0xFF) << 40 | ((long) in[offset + 3] & 0xFF) << 32 | ((long) in[offset + 4] & 0xFF) << 24 | ((long) in[offset + 5] & 0xFF) << 16 | ((long) in[offset + 6] & 0xFF) << 8 | ((long) in[offset + 7] & 0xFF));
        }
        else if (byteOrder == BYTE_ORDER_LITTLE_ENDIAN)
        {
            valueLong = (long) (((long) in[offset + 7] & 0xFF) << 56 | ((long) in[offset + 6] & 0xFF) << 48 | ((long) in[offset + 5] & 0xFF) << 40 | ((long) in[offset + 4] & 0xFF) << 32 | ((long) in[offset + 3] & 0xFF) << 24 | ((long) in[offset + 2] & 0xFF) << 16 | ((long) in[offset + 1] & 0xFF) << 8 | ((long) in[offset] & 0xFF));
        }
        else
        {
            throw new ByteInterpretationException(MESSAGE_UNRECOGNIZED_BYTE_ORDER);
        }
        
        return valueLong;
    }    
    
    public static float bytesToFloat(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        
        float valueFloat = 0;
        
        valueFloat = Float.intBitsToFloat(bytesToInt(in, offset, byteOrder));
        
        return valueFloat;
    }    
    
    public static double bytesToDouble(byte[] in, int offset, int byteOrder) throws ByteInterpretationException
    {
        
        double valueDouble = 0;
        
        valueDouble = Double.longBitsToDouble(bytesToLong(in, offset, byteOrder));
        
        return valueDouble;
    }
    
    public static int swapBytePositions(short value)
    {
        int valueInt = 0;
       
        valueInt = ((value & 0xFF) << 8) | ((value & 0xFF00) >> 8);
        
        return valueInt;        
    }
    
    /**
     * Use this function to interpret bytes as a C-Style NULL-terminated String. 
     * Note that the returned value will be a regular Java String object, and 
     * will <b>not</b> be NULL-terminated.
     */
    public static String bytesNullTerminatedToString(byte[] in, int offset) throws ByteInterpretationException
    {
        StringBuilder sb = new StringBuilder();
        for (int i = offset; i < in.length; ++i)
        {
            int ch = byteToUnsignedChar(in[i]);
            if (ch == 0)
            {
                break;
            }
            sb.append(((char) ch));
        }
        return sb.toString();
    }
    
    public static String bytesNullTerminatedToString(byte[] in, int offset, int maxLength) throws ByteInterpretationException
    {
        StringBuilder sb = new StringBuilder();
        for (int i = offset; i < in.length; ++i)
        {
            int ch = byteToUnsignedChar(in[i]);
            if (ch == 0 || sb.length() == maxLength)
            {
                break;
            }
            sb.append(((char) ch));
        }
        return sb.toString();
    } 

    public static String bytesToString(byte[] in, int offset, int length) throws ByteInterpretationException
    {
        StringBuilder sb = new StringBuilder();
        for (int i = offset; i < in.length; ++i)
        {
            int ch = byteToUnsignedChar(in[i]);
            sb.append(((char) ch));
            if (sb.length() == length)
            {
                break;
            }
        }
        return sb.toString();
    }    
    
    /**
     * Returns a hexadecimal representation of the input bytes. 
     * @param bytes Bytes to encode as a Hex String.
     * @param upperCaseHexString Pass true to get an upper case Hex String, false for lower case.
     * @return A hexadecimal representation of the input bytes.
     */
    public static String bytesToHexString(byte[] bytes, boolean upperCaseHexString)
    {
        StringBuilder sb = new StringBuilder(bytes.length);
        if (!upperCaseHexString)
        {
            for (int i = 0; i < bytes.length; ++i)
            {
                sb.append(String.format("%02x", bytes[i])); // Using upper case X instead of lower case x will result in the hex string being produced with upper case letters.
            }        
        }
        else
        {
            for (int i = 0; i < bytes.length; ++i)
            {
                sb.append(String.format("%02X", bytes[i]));
            }            
        }
        return sb.toString().trim();
    }    
    
}
