/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * CommonUtil.java
 *
 * Created on 19 July 2008, 20:18
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package net.infantai.util.common;


import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.HttpURLConnection;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.*;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class CommonUtil
{
    
    private static final int URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT = 30000; // milliseconds
    private static final int URL_CONNECTION_DEFAULT_READ_TIMEOUT = 30000; // milliseconds
    
    public CommonUtil() 
    {
        
    }
    
    public static String stringArrayToString(String[] strArray, String separator)
    {
        String strOutput = null;
        if (strArray == null)
        {
            return strOutput;
        }
        if (strArray.length < 1)
        {
            strOutput = "";
            return strOutput;
        }
        for (int i = 0; i < strArray.length; ++i)
        {
            if (strArray[i] == null || strArray[i].length() == 0)
            {
                continue;
            }
            if (strOutput == null)
            {
                strOutput = strArray[i];
            }
            else
            {
                strOutput += separator + strArray[i];
            }
        }
        return strOutput;
    }
    
    public static String stringListToString(List<String> strList, String separator)
    {
        String strOutput = null;
        if (strList == null)
        {
            return strOutput;
        }
        if (strList.size() < 1)
        {
            strOutput = "";
            return strOutput;
        }
        for (int i = 0; i < strList.size(); ++i)
        {
            if (strList.get(i) == null || strList.get(i).length() == 0)
            {
                continue;
            }
            if (strOutput == null)
            {
                strOutput = strList.get(i);
            }
            else
            {
                strOutput += separator + strList.get(i);
            }
        }
        return strOutput;
    }
    
    public static boolean charArrayContainsChar(char[] charArray, char ch)
    {
        for (int i = 0; i < charArray.length; ++i)
        {
            if (charArray[i] == ch)
            {
                return true;
            }
        }
        return false;
    }
    
    public static String trimCharacters(char[] charactersToTrim, String str)
    {
        String trimmedString = str;
        
        if (trimmedString == null || trimmedString.length() == 0 || charactersToTrim == null)
        {
            return trimmedString;
        }
        
        while (charArrayContainsChar(charactersToTrim, trimmedString.charAt(0)))
        {
            trimmedString = trimmedString.substring(1);
            if (trimmedString.length() == 0)
            {
                return trimmedString;
            }
        }
        
        if (trimmedString.length() == 0)
        {
            return trimmedString;
        }        
        
        while (charArrayContainsChar(charactersToTrim, trimmedString.charAt(trimmedString.length() - 1)))
        {
            trimmedString = trimmedString.substring(0, trimmedString.length() - 1);
            if (trimmedString.length() == 0)
            {
                return trimmedString;
            }            
        }
        return trimmedString.trim();
    }
    
    public static String trimCharacters(char[] charactersToTrim, String str, int maxCharactersToTrimLeft, int maxCharactersToTrimRight)
    {
        String trimmedString = str;
        
        if (trimmedString == null || trimmedString.length() == 0 || charactersToTrim == null)
        {
            return trimmedString;
        }
        
        int trimCountLeft = 0;
        while (charArrayContainsChar(charactersToTrim, trimmedString.charAt(0)))
        {
            trimmedString = trimmedString.substring(1);
            trimCountLeft++;
            if (trimCountLeft >= maxCharactersToTrimLeft)
            {
                break;
            }
            if (trimmedString.length() == 0)
            {
                return trimmedString;
            }
        }
        
        if (trimmedString.length() == 0)
        {
            return trimmedString;
        }        
        
        int trimCountRight = 0;
        while (charArrayContainsChar(charactersToTrim, trimmedString.charAt(trimmedString.length() - 1)))
        {
            trimmedString = trimmedString.substring(0, trimmedString.length() - 1);
            trimCountRight++;
            if (trimCountRight >= maxCharactersToTrimRight)
            {
                break;
            }
            if (trimmedString.length() == 0)
            {
                return trimmedString;
            }            
        }
        return trimmedString.trim();
    }    
    
    public static String trimCharactersLeft(char[] charactersToTrim, String str)
    {
        String trimmedString = str;
        
        if (trimmedString == null || trimmedString.length() == 0 || charactersToTrim == null)
        {
            return trimmedString;
        }
        
        while (charArrayContainsChar(charactersToTrim, trimmedString.charAt(0)))
        {
            trimmedString = trimmedString.substring(1);
            if (trimmedString.length() == 0)
            {
                break;
            }
        }
        
        return trimmedString.trim();
    }
    
    public static String trimCharactersRight(char[] charactersToTrim, String str)
    {
        String trimmedString = str;
        
        if (trimmedString == null || trimmedString.length() == 0 || charactersToTrim == null)
        {
            return trimmedString;
        }    
        
        while (charArrayContainsChar(charactersToTrim, trimmedString.charAt(trimmedString.length() - 1)))
        {
            trimmedString = trimmedString.substring(0, trimmedString.length() - 1);
            if (trimmedString.length() == 0)
            {
                return trimmedString;
            }            
        }
        return trimmedString.trim();
    }    

    public static boolean arrayContainsValue(char[] inputch, char c)
    {
        boolean found = false;
        for (int i = 0; i < inputch.length; ++i)
        {
            if (inputch[i] == c)
            {
                found = true;
                break;
            }
        }
        return found;
    }    
    
    public static Set<Character> stringToSetOfCharacters(String input)
    {
        Set<Character> chset = new LinkedHashSet<Character>();
        for (int i = 0; i < input.length(); ++i)
        {
            Character ch = new Character(input.charAt(i));
            chset.add(ch);
        }
        return chset;
    }    
    
    public static String getStackTraceAsString(Throwable ex)
    {
        String stackTrace = "";
        StringWriter stringWriter = new StringWriter();
        PrintWriter printWriter = new PrintWriter(stringWriter);
        ex.printStackTrace(printWriter);
        stackTrace = stringWriter.toString();
        return stackTrace;
    }
    
    public static FetchedFile fetch(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        URL url;
        InputStream urlStream = null;
        
        FetchedFile fetchedfile = new FetchedFile();
        
        // Make an attempt to open the URL connection.
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("http") || url.getProtocol().equalsIgnoreCase("https"))
            {
                HttpURLConnection urlConnection = null;
                urlConnection = (HttpURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setInstanceFollowRedirects(followRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);

                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;
            }
            else if (url.getProtocol().equalsIgnoreCase("ftp"))
            {
                URLConnection urlConnection = null;
                urlConnection = url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);

                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;                
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }                 
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        finally
        {
            if (urlStream != null)
            {
                try
                {
                    urlStream.close();
                }
                catch (IOException ioe)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                    ioe.printStackTrace();
                }
            }        
        }
        return fetchedfile;        
    }    
    
    public static FetchedFile fetch(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int connectTimeout, int readTimeout) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        URL url;
        InputStream urlStream = null;
        
        FetchedFile fetchedfile = new FetchedFile();
        
        // Make an attempt to open the URL connection.
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("http") || url.getProtocol().equalsIgnoreCase("https"))
            {
                HttpURLConnection urlConnection = null;
                urlConnection = (HttpURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(connectTimeout);
                urlConnection.setReadTimeout(readTimeout);
                urlConnection.setInstanceFollowRedirects(followRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);

                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;
            }
            else if (url.getProtocol().equalsIgnoreCase("ftp"))
            {
                URLConnection urlConnection = null;
                urlConnection = url.openConnection();
                urlConnection.setConnectTimeout(connectTimeout);
                urlConnection.setReadTimeout(readTimeout);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);

                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;                
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }                 
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        finally
        {
            if (urlStream != null)
            {
                try
                {
                    urlStream.close();
                }
                catch (IOException ioe)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                    ioe.printStackTrace();
                }
            }        
        }
        return fetchedfile;        
    }    
    
    public static FetchedFile fetch(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int connectTimeout, int readTimeout, int maxRetries) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        int retryWaitMS = 10000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetch(strFileURL, strUserAgent, maxFileSize, followRedirects, connectTimeout, readTimeout);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }
    }

    public static FetchedFile fetch(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int connectTimeout, int readTimeout, int maxRetries, int retryIntervalSeconds) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        
        int retryWaitMS = retryIntervalSeconds * 1000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetch(strFileURL, strUserAgent, maxFileSize, followRedirects, connectTimeout, readTimeout);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }
        
    }
    
    public static FetchedFile fetchWithPost(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, Map<String, String> formParameters) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        URL url;
        InputStream urlStream = null;
        
        FetchedFile fetchedfile = new FetchedFile();
        
        // Make an attempt to open the URL connection.
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("http") || url.getProtocol().equalsIgnoreCase("https"))
            {
                HttpURLConnection urlConnection = null;
                
                urlConnection = (HttpURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setRequestMethod("POST");
                urlConnection.setInstanceFollowRedirects(followRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setDoOutput(true);
                
                String queryString = "";
                boolean first = true;
                Iterator<Map.Entry<String, String>> formParametersIterator = formParameters.entrySet().iterator();
                while (formParametersIterator.hasNext())
                {
                    Map.Entry<String, String> formParameterObject = formParametersIterator.next();
                    String parameter = formParameterObject.getKey();
                    String parameterValue = formParameterObject.getValue();
                    
                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        queryString += "&";
                    }
                    
                    String urlEncodedParameter = URLEncoder.encode(parameter, "UTF-8");
                    String urlEncodedParameterValue = URLEncoder.encode(parameterValue, "UTF-8");
                    
                    queryString += urlEncodedParameter + "=" + urlEncodedParameterValue;
                    
                    
                }
                urlConnection.setAllowUserInteraction(false);

                OutputStream urlOutStream = urlConnection.getOutputStream();
                BufferedWriter buffWriter = new BufferedWriter(new OutputStreamWriter(urlOutStream, "UTF-8"));
                buffWriter.write(queryString);
                buffWriter.flush();
                buffWriter.close();
                urlOutStream.close();
                
                
                
                
                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;
            }
            
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }                 
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        finally
        {
            if (urlStream != null)
            {
                try
                {
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }
        return fetchedfile;      
    }
        
    /**
     * This method makes a GET request to the specified strFileURL, and returns 
     * the response as a FetchedFile object. Uses java.net.HttpURLConnection.
     * It tries up to maxRetries times, sleeping for 10 seconds between each attempt.
     * @param strFileURL The URL to which the request will be made.
     * @param strUserAgent The User-Agent String to be passed in the request header.
     * @param maxFileSize
     * @param followRedirects
     * @param maxRetries
     * @return FetchedFile
     * @throws UnsupportedFileSizeException
     * @throws SocketTimeoutException
     * @throws SocketException
     * @throws FileNotFoundException
     * @throws IOException
     * @throws InterruptedException 
     */
    public static FetchedFile fetch(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int maxRetries) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        
        int retryWaitMS = 10000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetch(strFileURL, strUserAgent, maxFileSize, followRedirects);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }
        
    }
    
    /**
     * This method makes a POST request to the specified strFileURL, and returns 
     * the response as a FetchedFile object. Uses java.net.HttpURLConnection.
     * It tries up to maxRetries times, sleeping for 10 seconds between each attempt.
     * @param strFileURL The URL to which the POST request will be made.
     * @param strUserAgent The User-Agent String to be passed in the request header.
     * @param maxFileSize
     * @param followRedirects
     * @param formParameters
     * @param maxRetries
     * @return FetchedFile
     * @throws UnsupportedFileSizeException
     * @throws SocketTimeoutException
     * @throws SocketException
     * @throws FileNotFoundException
     * @throws IOException
     * @throws InterruptedException 
     */
    public static FetchedFile fetchWithPost(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, Map<String, String> formParameters, int maxRetries) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        
        int retryWaitMS = 10000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetchWithPost(strFileURL, strUserAgent, maxFileSize, followRedirects, formParameters);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }        
        
    }
        
    
    public static FileBytesWithHTTPHeaderFields fetchFile(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int connectTimeout, int readTimeout, long lastModified, String eTag, boolean disconnect) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        FileBytesWithHTTPHeaderFields fbwri = null;
        URL url;
        InputStream urlStream = null;
        byte[] fileBytes = null;
        String strRedirLocation = null;
        URL redirURL = null;
        int httpResponseCode = -1;
        HttpURLConnection urlConnection = null;
        
        // Make an attempt to open the URL connection.
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("http") || url.getProtocol().equalsIgnoreCase("https"))
            {
                urlConnection = (HttpURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(connectTimeout);
                urlConnection.setReadTimeout(readTimeout);
                urlConnection.setInstanceFollowRedirects(followRedirects);
//                urlConnection.setRequestProperty("Connection", "keep-alive");
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                if (eTag != null)
                {
                    urlConnection.setRequestProperty("If-None-Match", eTag);
                }
                if (lastModified != -1)
                {
                    urlConnection.setIfModifiedSince(lastModified);
                }
                urlConnection.setAllowUserInteraction(false);
                
                urlStream = urlConnection.getInputStream();
                
                httpResponseCode = urlConnection.getResponseCode();
                
                if (httpResponseCode == HttpURLConnection.HTTP_NOT_MODIFIED)
                {
                    while (urlStream.read() != -1)
                    {
                        // needed to make sure that the underlying sockets are available as soon as possible
                    }
                    
                    return null;
                }
                if (httpResponseCode == HttpURLConnection.HTTP_MOVED_TEMP || httpResponseCode == HttpURLConnection.HTTP_MOVED_PERM || httpResponseCode == HttpURLConnection.HTTP_SEE_OTHER || httpResponseCode == 307) // 307 = Temporary Redirect, HTTP 1.1
                {
                    strRedirLocation = urlConnection.getHeaderField("Location");
                    
                    if (strRedirLocation != null)
                    {
                        redirURL = new java.net.URL(url, strRedirLocation);
                        System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "WARNING !! Found Redirect => " + strRedirLocation + " on " + strFileURL);
                        if (!followRedirects)
                        {
                            while (urlStream.read() != -1)
                            {
                                // needed to make sure that the underlying sockets are available as soon as possible
                            }                                
                            fbwri = new FileBytesWithHTTPHeaderFields();
                            fbwri.setFileBytes(null);
                            fbwri.setHttpResponseCode(httpResponseCode);                            
                            fbwri.setRedirectURL(redirURL.toExternalForm());
                            return fbwri;
                        }
                        
                        // HTTP to HTTP redirects are handled internally by HttpURLConnection. HTTP to HTTPS redirects have to be handled separately.
                        else if(redirURL.getProtocol().equalsIgnoreCase("https") && followRedirects)
                        {
                            return fetchFileHttps(redirURL.toExternalForm(), strUserAgent, maxFileSize, connectTimeout, readTimeout, lastModified, eTag, followRedirects, disconnect);
                        }
                        
                    }
                }
                    
                long lastModifiedFromHeader = urlConnection.getHeaderFieldDate("Last-Modified", -1);
                String eTagFromHeader = urlConnection.getHeaderField("ETag");               

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fileBytes = barrostream.toByteArray();
                barrostream.close();
                barrostream = null;                
                
                fbwri = new FileBytesWithHTTPHeaderFields();
                fbwri.setFileBytes(fileBytes);
                fbwri.setHttpResponseCode(httpResponseCode);
                fbwri.setLastModified(lastModifiedFromHeader);
                fbwri.setETag(eTagFromHeader);
                if (redirURL != null)
                {
                    fbwri.setRedirectURL(redirURL.toExternalForm());                
                }
                
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        finally
        {
            if (disconnect)
            {
                if (urlConnection != null)
                {
                    urlConnection.disconnect();
                }            
            }
            if (urlStream != null)
            {
                try
                {
//                    urlStream.skip(maxFileSize);
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }        
        return fbwri;
    }
    
    /**
     * Clone of fetchFile. 
     * This method throws EOFException instead of returning null, if the URLConnection's InputStream is empty.
     */
    public static FileBytesWithHTTPHeaderFields fetchFileEx(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int connectTimeout, int readTimeout, long lastModified, String eTag, boolean disconnect) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, EOFException
    {
        FileBytesWithHTTPHeaderFields fbwri = null;
        URL url;
        InputStream urlStream = null;
        byte[] fileBytes = null;
        String strRedirLocation = null;
        URL redirURL = null;
        int httpResponseCode = -1;
        HttpURLConnection urlConnection = null;
        
        // Make an attempt to open the URL connection.
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("http") || url.getProtocol().equalsIgnoreCase("https"))
            {
                urlConnection = (HttpURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(connectTimeout);
                urlConnection.setReadTimeout(readTimeout);
                urlConnection.setInstanceFollowRedirects(followRedirects);
//                urlConnection.setRequestProperty("Connection", "keep-alive");
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                if (eTag != null)
                {
                    urlConnection.setRequestProperty("If-None-Match", eTag);
                }
                if (lastModified != -1)
                {
                    urlConnection.setIfModifiedSince(lastModified);
                }
                urlConnection.setAllowUserInteraction(false);
                
                urlStream = urlConnection.getInputStream();
                
                httpResponseCode = urlConnection.getResponseCode();
                
                if (httpResponseCode == HttpURLConnection.HTTP_NOT_MODIFIED)
                {
                    while (urlStream.read() != -1)
                    {
                        // needed to make sure that the underlying sockets are available as soon as possible
                    }
                    
                    return null;
                }
                if (httpResponseCode == HttpURLConnection.HTTP_MOVED_TEMP || httpResponseCode == HttpURLConnection.HTTP_MOVED_PERM || httpResponseCode == HttpURLConnection.HTTP_SEE_OTHER || httpResponseCode == 307) // 307 = Temporary Redirect, HTTP 1.1
                {
                    strRedirLocation = urlConnection.getHeaderField("Location");
                    
                    if (strRedirLocation != null)
                    {
                        redirURL = new java.net.URL(url, strRedirLocation);
                        System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "WARNING !! Found Redirect => " + strRedirLocation + " on " + strFileURL);
                        if (!followRedirects)
                        {
                            while (urlStream.read() != -1)
                            {
                                // needed to make sure that the underlying sockets are available as soon as possible
                            }                                
                            fbwri = new FileBytesWithHTTPHeaderFields();
                            fbwri.setFileBytes(null);
                            fbwri.setHttpResponseCode(httpResponseCode);                            
                            fbwri.setRedirectURL(redirURL.toExternalForm());
                            return fbwri;
                        }
                        
                        // HTTP to HTTP redirects are handled internally by HttpURLConnection. HTTP to HTTPS redirects have to be handled separately.
                        else if(redirURL.getProtocol().equalsIgnoreCase("https") && followRedirects)
                        {
                            return fetchFileHttpsEx(redirURL.toExternalForm(), strUserAgent, maxFileSize, connectTimeout, readTimeout, lastModified, eTag, followRedirects, disconnect);
                        }
                        
                    }
                }
                    
                long lastModifiedFromHeader = urlConnection.getHeaderFieldDate("Last-Modified", -1);
                String eTagFromHeader = urlConnection.getHeaderField("ETag");              

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    throw new EOFException("Empty InputStream: " + strFileURL);
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fileBytes = barrostream.toByteArray();
                barrostream.close();
                barrostream = null;                
                
                fbwri = new FileBytesWithHTTPHeaderFields();
                fbwri.setFileBytes(fileBytes);
                fbwri.setHttpResponseCode(httpResponseCode);
                fbwri.setLastModified(lastModifiedFromHeader);
                fbwri.setETag(eTagFromHeader);
                if (redirURL != null)
                {
                    fbwri.setRedirectURL(redirURL.toExternalForm());                
                }
                
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        finally
        {
            if (disconnect)
            {
                if (urlConnection != null)
                {
                    urlConnection.disconnect();
                }            
            }
            if (urlStream != null)
            {
                try
                {
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }        
        return fbwri;
    }    
    
    public static FileBytesWithHTTPHeaderFields fetchFileHttps(String strFileURL, String strUserAgent, long maxFileSize, int connectTimeout, int readTimeout, long lastModified, String eTag, boolean followHttpsRedirects, boolean disconnect) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        FileBytesWithHTTPHeaderFields fbwri = null;
        URL url;
        InputStream urlStream = null;
        byte[] fileBytes = null;
        String strRedirLocation = null;
        URL redirURL = null;
        int httpResponseCode = -1;
        HttpsURLConnection urlConnection = null;
        
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("https"))
            {
                SSLContext context = SSLContext.getInstance("TLS");
                CustomTrustManager[] ctmArray = new CustomTrustManager[1];
                CustomTrustManager ctm = new CustomTrustManager();
                ctmArray[0] = ctm;
                context.init(null, ctmArray , null);
                
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setInstanceFollowRedirects(followHttpsRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                if (eTag != null)
                {
                    urlConnection.setRequestProperty("If-None-Match", eTag);
                }
                if (lastModified != -1)
                {
                    urlConnection.setIfModifiedSince(lastModified);
                }
                urlStream = urlConnection.getInputStream();
                
                httpResponseCode = urlConnection.getResponseCode();
                
                if (httpResponseCode == HttpURLConnection.HTTP_NOT_MODIFIED)
                {
                    while (urlStream.read() != -1)
                    {
                        // needed to make sure that the underlying sockets are available as soon as possible
                    }
                    
                    return null;
                }
                if (httpResponseCode == HttpURLConnection.HTTP_MOVED_TEMP || httpResponseCode == HttpURLConnection.HTTP_MOVED_PERM || httpResponseCode == HttpURLConnection.HTTP_SEE_OTHER || httpResponseCode == 307) // 307 = Temporary Redirect, HTTP 1.1
                {
                    strRedirLocation = urlConnection.getHeaderField("Location");
                    
                    if (strRedirLocation != null)
                    {
                        redirURL = new java.net.URL(url, strRedirLocation);
                        System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "WARNING !! Found Redirect => " + strRedirLocation + " on " + strFileURL);
                        if (!followHttpsRedirects)
                        {
                            while (urlStream.read() != -1)
                            {
                                // needed to make sure that the underlying sockets are available as soon as possible
                            }                                
                            fbwri = new FileBytesWithHTTPHeaderFields();
                            fbwri.setFileBytes(null);
                            fbwri.setHttpResponseCode(httpResponseCode);                            
                            fbwri.setRedirectURL(redirURL.toExternalForm());
                            return fbwri;
                        }
                    }
                }
                    
                long lastModifiedFromHeader = urlConnection.getHeaderFieldDate("Last-Modified", -1);
                String eTagFromHeader = urlConnection.getHeaderField("ETag");          

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fileBytes = barrostream.toByteArray();
                barrostream.close();
                barrostream = null;                
                
                fbwri = new FileBytesWithHTTPHeaderFields();
                fbwri.setFileBytes(fileBytes);
                fbwri.setHttpResponseCode(httpResponseCode);
                fbwri.setLastModified(lastModifiedFromHeader);
                fbwri.setETag(eTagFromHeader);
                if (redirURL != null)
                {
                    fbwri.setRedirectURL(redirURL.toExternalForm());                
                }
                
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        catch (NoSuchAlgorithmException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        catch (KeyManagementException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        finally
        {
            if (disconnect)
            {
                if (urlConnection != null)
                {
                    urlConnection.disconnect();
                }            
            }
            if (urlStream != null)
            {
                try
                {
//                    urlStream.skip(maxFileSize);
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }        
        return fbwri;

    }
    
    /**
     * Same as fetchFileHttps, except that this method throws EOFException 
     * instead of returning null, if the URLConnection's InputStream is empty.
     */
    public static FileBytesWithHTTPHeaderFields fetchFileHttpsEx(String strFileURL, String strUserAgent, long maxFileSize, int connectTimeout, int readTimeout, long lastModified, String eTag, boolean followHttpsRedirects, boolean disconnect) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, EOFException
    {
        FileBytesWithHTTPHeaderFields fbwri = null;
        URL url;
        InputStream urlStream = null;
        byte[] fileBytes = null;
        String strRedirLocation = null;
        URL redirURL = null;
        int httpResponseCode = -1;
        HttpsURLConnection urlConnection = null;

        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("https"))
            {
                SSLContext context = SSLContext.getInstance("TLS");
                CustomTrustManager[] ctmArray = new CustomTrustManager[1];
                CustomTrustManager ctm = new CustomTrustManager();
                ctmArray[0] = ctm;
                context.init(null, ctmArray, null);
                
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setInstanceFollowRedirects(followHttpsRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                if (eTag != null)
                {
                    urlConnection.setRequestProperty("If-None-Match", eTag);
                }
                if (lastModified != -1)
                {
                    urlConnection.setIfModifiedSince(lastModified);
                }
                urlStream = urlConnection.getInputStream();
                
                httpResponseCode = urlConnection.getResponseCode();
                
                if (httpResponseCode == HttpURLConnection.HTTP_NOT_MODIFIED)
                {
                    while (urlStream.read() != -1)
                    {
                        // needed to make sure that the underlying sockets are available as soon as possible
                    }
                    
                    return null;
                }
                if (httpResponseCode == HttpURLConnection.HTTP_MOVED_TEMP || httpResponseCode == HttpURLConnection.HTTP_MOVED_PERM || httpResponseCode == HttpURLConnection.HTTP_SEE_OTHER || httpResponseCode == 307) // 307 = Temporary Redirect, HTTP 1.1
                {
                    strRedirLocation = urlConnection.getHeaderField("Location");
                    
                    if (strRedirLocation != null)
                    {
                        redirURL = new java.net.URL(url, strRedirLocation);
                        System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "WARNING !! Found Redirect => " + strRedirLocation + " on " + strFileURL);
                        if (!followHttpsRedirects)
                        {
                            while (urlStream.read() != -1)
                            {
                                // needed to make sure that the underlying sockets are available as soon as possible
                            }                                
                            fbwri = new FileBytesWithHTTPHeaderFields();
                            fbwri.setFileBytes(null);
                            fbwri.setHttpResponseCode(httpResponseCode);                            
                            fbwri.setRedirectURL(redirURL.toExternalForm());
                            return fbwri;
                        }
                    }
                }
                    
                long lastModifiedFromHeader = urlConnection.getHeaderFieldDate("Last-Modified", -1);
                String eTagFromHeader = urlConnection.getHeaderField("ETag");      

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    throw new EOFException("Empty InputStream: " + strFileURL);
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fileBytes = barrostream.toByteArray();
                barrostream.close();
                barrostream = null;                
                
                fbwri = new FileBytesWithHTTPHeaderFields();
                fbwri.setFileBytes(fileBytes);
                fbwri.setHttpResponseCode(httpResponseCode);
                fbwri.setLastModified(lastModifiedFromHeader);
                fbwri.setETag(eTagFromHeader);
                if (redirURL != null)
                {
                    fbwri.setRedirectURL(redirURL.toExternalForm());                
                }
                
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        catch (NoSuchAlgorithmException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        catch (KeyManagementException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        finally
        {
            if (disconnect)
            {
                if (urlConnection != null)
                {
                    urlConnection.disconnect();
                }            
            }
            if (urlStream != null)
            {
                try
                {
//                    urlStream.skip(maxFileSize);
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }        
        return fbwri;

    }
    
    public static FetchedFile fetchHttps(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        URL url;
        InputStream urlStream = null;
        
        FetchedFile fetchedfile = new FetchedFile();
        
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("https"))
            {
                
                SSLContext context = SSLContext.getInstance("TLS");
                CustomTrustManager[] ctmArray = new CustomTrustManager[1];
                CustomTrustManager ctm = new CustomTrustManager();
                ctmArray[0] = ctm;
                context.init(null, ctmArray, null);
                
                HttpsURLConnection urlConnection = null;
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setInstanceFollowRedirects(followRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                
                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }                 
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        catch (NoSuchAlgorithmException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        catch (KeyManagementException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        finally
        {
            if (urlStream != null)
            {
                try
                {
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }
        return fetchedfile;        
    }
    
    public static FetchedFile fetchHttps(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int maxRetries) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        
        int retryWaitMS = 10000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetchHttps(strFileURL, strUserAgent, maxFileSize, followRedirects);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }                

    }
    
    public static FetchedFile fetchHttps(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, int maxRetries, int retryIntervalSeconds) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        
        int retryWaitMS = retryIntervalSeconds * 1000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetchHttps(strFileURL, strUserAgent, maxFileSize, followRedirects);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }

    }
    
    public static FetchedFile fetchHttpsWithPost(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, Map<String, String> formParameters) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException
    {
        URL url;
        InputStream urlStream = null;
        
        FetchedFile fetchedfile = new FetchedFile();
        
        // Make an attempt to open the URL connection.
        try
        {
            url = new URL(strFileURL);
            
            if (url.getProtocol().equalsIgnoreCase("https"))
            {
                
                SSLContext context = SSLContext.getInstance("TLS");
                CustomTrustManager[] ctmArray = new CustomTrustManager[1];
                CustomTrustManager ctm = new CustomTrustManager();
                ctmArray[0] = ctm;
                context.init(null, ctmArray, null);
                
                HttpsURLConnection urlConnection = null;
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setConnectTimeout(URL_CONNECTION_DEFAULT_CONNECT_TIMEOUT);
                urlConnection.setReadTimeout(URL_CONNECTION_DEFAULT_READ_TIMEOUT);
                urlConnection.setRequestMethod("POST");
                urlConnection.setInstanceFollowRedirects(followRedirects);
                urlConnection.setRequestProperty("User-Agent", strUserAgent);
                urlConnection.setAllowUserInteraction(false);
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setDoOutput(true);
                
                
                String queryString = "";
                boolean first = true;
                Iterator<Map.Entry<String, String>> formParametersIterator = formParameters.entrySet().iterator();
                while (formParametersIterator.hasNext())
                {
                    Map.Entry<String, String> formParameterObject = formParametersIterator.next();
                    String parameter = formParameterObject.getKey();
                    String parameterValue = formParameterObject.getValue();
                    
                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        queryString += "&";
                    }
                    
                    String urlEncodedParameter = URLEncoder.encode(parameter, "UTF-8");
                    String urlEncodedParameterValue = URLEncoder.encode(parameterValue, "UTF-8");
                    
                    queryString += urlEncodedParameter + "=" + urlEncodedParameterValue;
                    
                    
                }
                
                OutputStream urlOutStream = urlConnection.getOutputStream();
                BufferedWriter buffWriter = new BufferedWriter(new OutputStreamWriter(urlOutStream, "UTF-8"));
                buffWriter.write(queryString);
                buffWriter.flush();
                buffWriter.close();
                urlOutStream.close();
                
                urlStream = urlConnection.getInputStream();

                // Try to determine the size of the file to be downloaded.
                long contentLength = (long) urlConnection.getContentLength();
                if (maxFileSize != -1)
                {
                    if (contentLength > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("Content-Length = " + contentLength + " !");
                    }
                }

                ByteArrayOutputStream barrostream = new ByteArrayOutputStream(10240);
                int totalBytes = 0;
                byte buffer[] = new byte[2048]; // A 2 KB Buffer
                int iBytes = urlStream.read(buffer);
                if (iBytes == -1)
                {
                    return null;	
                }
                barrostream.write(buffer, 0, iBytes);
                totalBytes += iBytes;

                while (iBytes != -1)
                {
                    iBytes = urlStream.read(buffer);
                    if (iBytes == -1)
                    {
                        break;	
                    }
                    barrostream.write(buffer, 0, iBytes);
                    totalBytes += iBytes;
                    if (maxFileSize != -1 && totalBytes > maxFileSize)
                    {
                        throw new UnsupportedFileSizeException("File size exceeds " + maxFileSize + " bytes !");
                    }
                }

                fetchedfile.setFileURL(urlConnection.getURL().toExternalForm());
                fetchedfile.setFileBytes(barrostream.toByteArray());
                fetchedfile.setFetchTime(System.currentTimeMillis());

                barrostream.close();
                barrostream = null;                
                urlConnection = null;
            }
        }
        catch (SocketTimeoutException ste)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ste.getMessage());
            System.err.println("[Socket Timeout] " + strFileURL);
            throw ste;
        }
        catch (SocketException se)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
                System.err.println("[Connection Error] " + strFileURL);
            }                 
            throw se;
        }
        catch (FileNotFoundException fnfe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
                System.err.println("[File Not Found] " + strFileURL);
            }
            throw fnfe;
        }
        catch (IOException ioe)
        {
            if (!strFileURL.endsWith("/robots.txt"))
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                System.err.println("[Server / Connection Error Or Access Denied] " + strFileURL);
            }
            throw ioe;
        }
        catch (NoSuchAlgorithmException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        catch (KeyManagementException kmex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + kmex.getMessage());
            kmex.printStackTrace();
        }
        finally
        {
            if (urlStream != null)
            {
                try
                {
                    urlStream.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                    ioex.printStackTrace();
                }
            }        
        }
        return fetchedfile;        
    }  
    
    public static FetchedFile fetchHttpsWithPost(String strFileURL, String strUserAgent, long maxFileSize, boolean followRedirects, Map<String, String> formParameters, int maxRetries) throws UnsupportedFileSizeException, SocketTimeoutException, SocketException, FileNotFoundException, IOException, InterruptedException
    {
        
        int retryWaitMS = 10000;
        int attempts = 0;
        while (true)
        {
            try
            {
                return fetchHttpsWithPost(strFileURL, strUserAgent, maxFileSize, followRedirects, formParameters);
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                attempts++;
                if (attempts <= maxRetries)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Attempting page fetch again (Retry " + attempts + ") after " + (retryWaitMS / 1000.0) + " seconds: " + strFileURL);
                    Thread.sleep(retryWaitMS);
                }
                else
                {
                    throw ioex;
                }
            }
        }        
        
    }
    
}