/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * TextUtil.java
 *
 * Created on 16 June 2019, 23:44
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package net.infantai.util.common;

import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Pattern;
import net.infantai.core.MathLib;
import net.infantai.nlp.ner.EntityTaggerOutput;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class TextUtil
{
    
    private Pattern patRangeTokenSplitter = null;
    
    public TextUtil()
    {
        patRangeTokenSplitter = Pattern.compile("\\W+");
    }
    
    public TextUtil(String defaultDelimiterRegexPattern)
    {
        patRangeTokenSplitter = Pattern.compile(defaultDelimiterRegexPattern);
    }
    
//    public String getNTokensFromPrevText(String prevText, int n)
//    {
//        String modifiedPrevText = "";
//        String[] prevTokens = patRangeTokenSplitter.split(prevText.trim());
//        if (prevTokens.length <= n)
//        {
//            modifiedPrevText = prevText;
//        }
//        else
//        {
//            for (int i = (prevTokens.length - n); i < prevTokens.length; ++i)
//            {
//                if (modifiedPrevText.length() == 0)
//                {
//                    modifiedPrevText = prevTokens[i];
//                }
//                else
//                {
//                    modifiedPrevText = modifiedPrevText + " " + prevTokens[i];
//                }
//            }
//        }
//        return modifiedPrevText.trim();
//    }   
    
    public String getNTokensFromPrevText(String prevText, int n)
    {
        String modifiedPrevText = "";
        String[] prevTokens = patRangeTokenSplitter.split(prevText.trim());
        for (int i = (prevTokens.length - n); i < prevTokens.length; ++i)
        {
            if (i < 0)
            {
                i = -1; // will be incremented to 0 in the next iteration
                continue;
            }
            if (modifiedPrevText.length() == 0)
            {
                modifiedPrevText = prevTokens[i];
            }
            else
            {
                modifiedPrevText = modifiedPrevText + " " + prevTokens[i];
            }
        }
        return modifiedPrevText.trim();
    }    

//    public String getNTokensFromNextText(String nextText, int n)
//    {
//        String modifiedNextText = "";
//        String[] nextTokens = patRangeTokenSplitter.split(nextText.trim());
//        if (nextTokens.length <= n)
//        {
//            modifiedNextText = nextText;
//        }
//        else
//        {
//            for (int i = 0; i < n; ++i)
//            {
//                if (modifiedNextText.length() == 0)
//                {
//                    modifiedNextText = nextTokens[i];
//                }
//                else
//                {
//                    modifiedNextText = modifiedNextText + " " + nextTokens[i];
//                }
//            }
//        }
//        return modifiedNextText.trim();
//    }        
    
    public String getNTokensFromNextText(String nextText, int n)
    {
        String modifiedNextText = "";
        String[] nextTokens = patRangeTokenSplitter.split(nextText.trim());
        for (int i = 0; i < n; ++i)
        {
            if (i < nextTokens.length)
            {
                if (modifiedNextText.length() == 0)
                {
                    modifiedNextText = nextTokens[i];
                }
                else
                {
                    modifiedNextText = modifiedNextText + " " + nextTokens[i];
                }
            }
            else
            {
                break;
            }
        }
        return modifiedNextText.trim();
    }            
    
    /**
     * Combines tokens in reverse order (right to left).
     * @param prevText
     * @param n
     * @return 
     */
    public Set<String> getUpToNTokensFromPrevText(String prevText, int n)
    {
        Set<String> nTokenWords = new LinkedHashSet<String>();
        String modifiedPrevText = "";
        String[] prevTokens = patRangeTokenSplitter.split(prevText.trim());
        int lbound = n < prevTokens.length ? prevTokens.length - n : 0;
        for (int i = (prevTokens.length - 1); i >= lbound; --i)
        {
            if (modifiedPrevText.length() == 0)
            {
                modifiedPrevText = prevTokens[i];
            }
            else
            {
                modifiedPrevText = prevTokens[i] + " " + modifiedPrevText;
            }
            nTokenWords.add(modifiedPrevText.trim());
        }
        return nTokenWords;
    }    
    
    /**
     * Combines tokens from left to right.
     * @param nextText
     * @param n
     * @return 
     */
    public Set<String> getUpToNTokensFromNextText(String nextText, int n)
    {
        Set<String> nTokenWords = new LinkedHashSet<String>();
        String modifiedNextText = "";
        String[] nextTokens = patRangeTokenSplitter.split(nextText.trim());
        int ubound = n < nextTokens.length ? n : nextTokens.length;
        for (int i = 0; i < ubound; ++i)
        {
            if (modifiedNextText.length() == 0)
            {
                modifiedNextText = nextTokens[i];
            }
            else
            {
                modifiedNextText = modifiedNextText + " " + nextTokens[i];
            }
            nTokenWords.add(modifiedNextText.trim());
        }
        return nTokenWords;
    }        
    
    /**
     * Combines tokens in reverse order (right to left).
     * @param prevText
     * @param n
     * @return 
     */
    public Set<String> getUpToNTokensFromPrevTextWithSingleTokens(String prevText, int n)
    {
        Set<String> nTokenWords = new LinkedHashSet<String>();
        String modifiedPrevText = "";
        String[] prevTokens = patRangeTokenSplitter.split(prevText.trim());
        int lbound = n < prevTokens.length ? prevTokens.length - n : 0;
        for (int i = (prevTokens.length - 1); i >= lbound; --i)
        {
            if (modifiedPrevText.length() == 0)
            {
                modifiedPrevText = prevTokens[i];
            }
            else
            {
                modifiedPrevText = prevTokens[i] + " " + modifiedPrevText;
            }
            nTokenWords.add(prevTokens[i]);
            nTokenWords.add(modifiedPrevText.trim());
        }
        return nTokenWords;
    }    
    
    /**
     * Combines tokens from left to right.
     * @param nextText
     * @param n
     * @return 
     */
    public Set<String> getUpToNTokensFromNextTextWithSingleTokens(String nextText, int n)
    {
        Set<String> nTokenWords = new LinkedHashSet<String>();
        String modifiedNextText = "";
        String[] nextTokens = patRangeTokenSplitter.split(nextText.trim());
        int ubound = n < nextTokens.length ? n : nextTokens.length;
        for (int i = 0; i < ubound; ++i)
        {
            if (modifiedNextText.length() == 0)
            {
                modifiedNextText = nextTokens[i];
            }
            else
            {
                modifiedNextText = modifiedNextText + " " + nextTokens[i];
            }
            nTokenWords.add(nextTokens[i]);
            nTokenWords.add(modifiedNextText.trim());
        }
        return nTokenWords;
    }    
    
    /**
     * This method assumes that the input List etoList is sorted (in ascending
     * order) by the index of the TokenGroup.
     *
     * @param etoList
     * @param etoIndexInList
     * @return
     */
    public int getPrevIndex(String title, List<EntityTaggerOutput> etoList, int etoIndexInList, int characterDistance)
    {
        int prevIndex = 0;
        EntityTaggerOutput eto = etoList.get(etoIndexInList);
        if (etoIndexInList > 0)
        {
            boolean foundBreakingCondition = false;
            for (int i = (etoIndexInList - 1); i >= 0; --i)
            {
                if (etoList.get(i).tg.getIndex() < eto.tg.getIndex())
                {
                    prevIndex = etoList.get(i).tg.getIndex() + etoList.get(i).tg.getSize();
                    if (i < (etoList.size() - 1))
                    {
//                        if (prevIndex >= etoList.get(i + 1).tg.getIndex())
                        int nextDistinctEntityTokenIndex = 0;
                        for (int j = (i + 1); j <= etoIndexInList; ++j)
                        {
                            if (etoList.get(i).tg.getIndex().intValue() == etoList.get(j).tg.getIndex().intValue())
                            {
                                continue;
                            }
                            nextDistinctEntityTokenIndex = etoList.get(j).tg.getIndex();
                        }
                        
                        if (prevIndex >= nextDistinctEntityTokenIndex)
                        {
                            continue;
                        }                        
                    }
                    if ((eto.tg.getIndex() - prevIndex) >= characterDistance)
                    {
                        foundBreakingCondition = true;
                        break;
                    }
                    else
                    {
                        if (etoList.get(i).isDictionaryMatch)
                        {
                            foundBreakingCondition = true;
                            break;
                        }
                    }
                }
            }
            if (!foundBreakingCondition)
            {
                prevIndex = 0;
            }
        }
        return prevIndex;
    }    

    /**
     * This method assumes that the input List etoList is sorted (in ascending
     * order) by the index of the TokenGroup.
     *
     * @param etoList
     * @param etoIndexInList
     * @return
     */
    public int getNextGreaterIndex(String title, List<EntityTaggerOutput> etoList, int etoIndexInList, int characterDistance)
    {
        int nextIndex = -1;
        EntityTaggerOutput eto = etoList.get(etoIndexInList);
        if (etoIndexInList < (etoList.size() - 1))
        {
            boolean foundBreakingCondition = false;
            for (int i = (etoIndexInList + 1); i < etoList.size(); ++i)
            {
                if (etoList.get(i).tg.getIndex() > eto.tg.getIndex())
                {
                    nextIndex = etoList.get(i).tg.getIndex();
                    if (nextIndex - (eto.tg.getIndex() + eto.tg.getSize()) >= characterDistance)
                    {
                        foundBreakingCondition = true;
                        break;
                    }
                    else
                    {
                        if (etoList.get(i).isDictionaryMatch)
                        {
                            foundBreakingCondition = true;
                            break;
                        }
                    }
                }
            }
            if (!foundBreakingCondition)
            {
                nextIndex = title.length();
            }
        }
        return nextIndex;
    }    
    
    public static int getNextIndexOfCharacter(String input, int startIndex, char[] targetCharacters)
    {
        int nextChIndex = -1;
        for (int i = startIndex; i < input.length(); ++i)
        {
            if (CommonUtil.arrayContainsValue(targetCharacters, input.charAt(i)))
            {
                nextChIndex = i;
                break;
            }
        }
        return nextChIndex;
    }
    
    public static int getPrevIndexOfCharacter(String input, int startIndex, char[] targetCharacters)
    {
        int prevChIndex = -1;
        for (int i = startIndex; i >= 0; --i)
        {
            if (CommonUtil.arrayContainsValue(targetCharacters, input.charAt(i)))
            {
                prevChIndex = i;
                break;
            }
        }
        return prevChIndex;
    }    
    
    public static Set<TokenGroup> getAllDistinctTokens(String input, List<String> delimiters, int maxTokenCount)
    {
        Set<TokenGroup> distinctTokens = new TreeSet<>();
        Tokenizer tokenizer = new Tokenizer();
        for (int j = 0; j < delimiters.size(); ++j)
        {
            distinctTokens.addAll(tokenizer.getAllTokens(input, delimiters.get(j), maxTokenCount));
        }
        return distinctTokens;
    }        
    
    public String discardTokens(String input, Set<String> tokensToDiscard)
    {
        StringBuilder sb = new StringBuilder();
        String[] parts = patRangeTokenSplitter.split(input);
        for (int i = 0; i < parts.length; ++i)
        {
            if (!tokensToDiscard.contains(parts[i]))
            {
                sb.append(parts[i]);
                if (i < (parts.length - 1))
                {
                    sb.append(' ');
                }
            }
        }
        return sb.toString().trim();
    }

    /*
    public String trimTokens(String input, Set<String> tokensToTrim)
    {
        StringBuilder sb = new StringBuilder();
        String[] parts = patRangeTokenSplitter.split(input);
        boolean doneChecking = false;
        
        // LTRIM
        for (int i = 0; i < parts.length; ++i)
        {
            if (!doneChecking && tokensToTrim.contains(parts[i]))
            {
                continue;
            }
            else
            {
                doneChecking = true;
            }
            sb.append(parts[i]);
            if (i < (parts.length - 1))
            {
                sb.append(' ');
            }
        }
        
        parts = patRangeTokenSplitter.split(sb.toString().trim());
        sb = new StringBuilder();
        doneChecking = false;
        
        // RTRIM
        int lastIndexToAppend = 0;
        for (int i = (parts.length - 1); i >= 0; --i)
        {
            if (tokensToTrim.contains(parts[i]))
            {
                continue;
            }
            lastIndexToAppend = i;
            break;
        }
        for (int i = 0; i <= lastIndexToAppend; ++i)
        {
            sb.append(parts[i]);
            if (i < lastIndexToAppend)
            {
                sb.append(' ');
            }        
        }
        return sb.toString().trim();
    }
    */    
    
    public String trimTokens(String input, Set<String> tokensToTrim)
    {
        StringBuilder sb = new StringBuilder();
        String[] parts = patRangeTokenSplitter.split(input);
        
        // LTRIM
        int firstIndexToAppend = -1;
        for (int i = 0; i < parts.length; ++i)
        {
            if (tokensToTrim.contains(parts[i]))
            {
                continue;
            }
            firstIndexToAppend = i;
            break;
        }
        
        // RTRIM
        int lastIndexToAppend = -1;
        if (firstIndexToAppend != -1) // if firstIndexToAppend is -1, it means that all the terms in the input text need to be discarded; no need for the reverse loop next
        {
            for (int i = (parts.length - 1); i >= 0; --i)
            {
                if (tokensToTrim.contains(parts[i]))
                {
                    continue;
                }
                lastIndexToAppend = i;
                break;
            }        
        }
        
        if (firstIndexToAppend != -1 && lastIndexToAppend != -1)
        {
            for (int i = firstIndexToAppend; i <= lastIndexToAppend; ++i)
            {
                sb.append(parts[i]);
                if (i < lastIndexToAppend)
                {
                    sb.append(' ');
                }        
            }        
        }
        return sb.toString().trim();
    }    
    
    public String getFirstToken(String input)
    {
        String firstToken = null;
        String[] parts = patRangeTokenSplitter.split(input);
        if (parts.length > 0) // make sure array has at least one element
        {
            firstToken = parts[0];
        }
        return firstToken;
    }    
    
    /**
     * "Logically Consistent" as input, will return "Consistent Logically" as output. 
     * Will work only with 2-Token strings.
     * @param input
     * @param pivots
     * @return 
     */
    public String getTokensTransposed(String input)
    {
        String sTransposed = null;
        if (input == null)
        {
            return input;
        }
        String[] parts = patRangeTokenSplitter.split(input);
        if (parts.length == 2)
        {
            sTransposed = parts[1] + " " + parts[0];
            return sTransposed;
        }
        return input;
    }    
    
    /**
     * "Sea Of Japan" as input, will return "Japan Sea" as output. The pivot term itself will not be included in the output.
     * @param input
     * @param pivots
     * @return 
     */
    public String getTokensTransposedOnPivot(String input, Set<String> pivots)
    {
        StringBuilder sb1 = new StringBuilder();
        StringBuilder sb2 = new StringBuilder();
        String[] parts = patRangeTokenSplitter.split(input);
        boolean foundPivot = false;
        if (parts.length >= 3)
        {
            for (int i = 0; i < parts.length; ++i)
            {
                if (!foundPivot)
                {
                    if (!pivots.contains(parts[i]))
                    {
                        sb1.append(parts[i]);
                        if (i < (parts.length - 1))
                        {
                            sb1.append(' ');
                        }
                    }
                    else
                    {
                        foundPivot = true;
                    }
                }
                else
                {
                    sb2.append(parts[i]);
                    if (i < (parts.length - 1))
                    {
                        sb2.append(' ');
                    }                                
                }
            }
            return sb2.toString().trim() + " " + sb1.toString().trim();
        }
        return input;
    }        
    
    public int getTokenCount(String input)
    {
        return patRangeTokenSplitter.split(input).length;
    }    
    
    public Map<String, Double> getTermFrequencyVector(String input, boolean normalize)
    {
        Map<String, Double> termFreq = new LinkedHashMap<String, Double>();
        String[] tokens = patRangeTokenSplitter.split(input.trim());
        for (int i = 0; i < tokens.length; ++i)
        {
            String token = tokens[i];
            Double freq = termFreq.get(token);
            if (freq == null)
            {
                freq = Double.valueOf(0.0);
            }
            freq++;
            termFreq.put(token, freq);
        }
        if (normalize)
        {
            termFreq = MathLib.normalizeTermFrequencyVector(termFreq);
        }
        return termFreq;
    }
    
    public static boolean hasOverlap(List<TokenBoundary> boundaries, int startIndex, int size)
    {
        boolean overlap = false;
        for (int i = 0; i < boundaries.size(); ++i) // expected size of the list is too small to consider using a binary search 
        {
            TokenBoundary tokenBoundary = boundaries.get(i);
            if ((startIndex >= tokenBoundary.startIndex && startIndex <= tokenBoundary.endIndex) || ((startIndex + size) >= tokenBoundary.startIndex && (startIndex + size) <= tokenBoundary.endIndex))
            {
                overlap = true;
                break;
            }
        }
        return overlap;
    }    
    
    public static void main(String[] args)
    {
        
        TextUtil tu = new TextUtil();
        long t1 = System.nanoTime();
        Set<String> words = tu.getUpToNTokensFromPrevText("one two three four five six", 5);
        long t2 = System.nanoTime();
        System.out.println((t2 - t1) / 1000000.0 + " ms.");
        for (String s : words)
        {
            System.out.println(s);
        }
        
        Map<String, Double> tf = tu.getTermFrequencyVector("The quick brown fox jumps over the lazy dog.".toLowerCase(), false);
        for (Map.Entry<String, Double> mapEntry : tf.entrySet())
        {
            System.out.println(mapEntry.getKey() + ", " + mapEntry.getValue());
        }
        
        Set<String> pivots = new HashSet<String>();
        pivots.add("of");
        pivots.add("in");
        System.out.println(tu.getTokensTransposedOnPivot("sea of japan", pivots));
        
        String inputForTokenTrimming = "four one two three";
        Set<String> tokensToTrim = new LinkedHashSet<>();
        tokensToTrim.add("one");
        tokensToTrim.add("two");
        tokensToTrim.add("seven");
        tokensToTrim.add("eight");
        String trimmedOutput = tu.trimTokens(inputForTokenTrimming, tokensToTrim);
        System.out.println(trimmedOutput);
        
    }
    
}
