/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.util.common;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class Tokenizer 
{
    
    private static final String TOKEN_DELIM_1 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".“”()[]{}`~|'‘’_‒–—\\/?=!+#*^&%©®™"; // splits on all common special characters

    public Tokenizer()
    {
        
    }
    
    public List<TokenGroup> getAllTokens(String textToTokenize, String delim, int maxTokensInGroup)
    {
        List<String> tokens = new ArrayList<String>(); 
        List<Integer> tokenIndex = new ArrayList<Integer>();
        List<Integer> tokenEndIndex = new ArrayList<Integer>();
        List<TokenGroup> alltokens = new ArrayList<TokenGroup>(); // key = token, value = token index
//        Tokenizer st = new Tokenizer(textToTokenize, delim, false); // can consider using a regex split here
        
        
        // custom tokenization routine
        char[] delimch = delim.toCharArray();
        char[] inputch = textToTokenize.toCharArray();
        
//        List<String> tokens = new ArrayList<String>();
        
        int startIndex = -1;
        
        for (int i = 0; i < inputch.length; ++i)
        {
            if (CommonUtil.arrayContainsValue(delimch, inputch[i]))
            {
                if (startIndex == -1)
                {
                    continue;
                }
                else
                {
                    tokenIndex.add(new Integer(startIndex));
                    tokenEndIndex.add(new Integer(i));
                    tokens.add(textToTokenize.substring(startIndex, i));
                    startIndex = -1;
                }                
            }
            else
            {
                if (startIndex == -1)
                {
                    startIndex = i;
                }
            }
        }
        
        if (startIndex != -1) // get the last token
        {
            tokenIndex.add(new Integer(startIndex));
            tokenEndIndex.add(new Integer(inputch.length));
            tokens.add(textToTokenize.substring(startIndex, inputch.length));
        }
        
        // Collect groups of 1, 2, 3, 4, 5, ... n tokens.
        int n = maxTokensInGroup;
        
        for (int k = n; k >= 1; --k)
        {
            
            for (int i = 0; i < tokens.size(); ++i)
            {
                if (i < (tokens.size() - (k - 1)))
                {
                    StringBuilder sb = new StringBuilder();
                    sb.append(tokens.get(i));
                    
                    int endIndex = tokenEndIndex.get(i).intValue();

                    for (int j = 1; j < k; ++j)
                    {
                        sb.append(' ');
                        sb.append(tokens.get(i + j));
                        
                        endIndex = tokenEndIndex.get(i + j).intValue();
                    }
                    
                    String token = sb.toString();
                    
//                    System.out.println("=> " + token);
                    
                    int tindex = tokenIndex.get(i).intValue();
                    
                    alltokens.add(new TokenGroup(token, tindex, (endIndex - tindex)));
                }
            }        
        }        
        
        return alltokens;
    }   
    
    public Set<String> getDistinctTokens(String textToTokenize, String delim, int maxTokensInGroup)
    {
        List<String> tokens = new ArrayList<String>(); 
        List<Integer> tokenIndex = new ArrayList<Integer>();
        List<Integer> tokenEndIndex = new ArrayList<Integer>();
        Set<String> distinctTokens = new LinkedHashSet<String>();
        
        
        // custom tokenization routine
        char[] delimch = delim.toCharArray();
        char[] inputch = textToTokenize.toCharArray();
        
//        List<String> tokens = new ArrayList<String>();
        
        int startIndex = -1;
        
        for (int i = 0; i < inputch.length; ++i)
        {
            if (CommonUtil.arrayContainsValue(delimch, inputch[i]))
            {
                if (startIndex == -1)
                {
                    continue;
                }
                else
                {
                    tokenIndex.add(new Integer(startIndex));
                    tokenEndIndex.add(new Integer(i));
                    tokens.add(textToTokenize.substring(startIndex, i));
                    startIndex = -1;
                }                
            }
            else
            {
                if (startIndex == -1)
                {
                    startIndex = i;
                }
            }
        }
        
        if (startIndex != -1) // get the last token
        {
            tokenIndex.add(new Integer(startIndex));
            tokenEndIndex.add(new Integer(inputch.length));
            tokens.add(textToTokenize.substring(startIndex, inputch.length));
        }
        
        // end of custom tokenization routine
        
        
        // Collect groups of 1, 2, 3, 4, 5, and 6 tokens.
        int n = maxTokensInGroup;
        
        for (int k = n; k >= 1; --k)
        {
            
            for (int i = 0; i < tokens.size(); ++i)
            {
                if (i < (tokens.size() - (k - 1)))
                {
                    StringBuilder sb = new StringBuilder();
                    sb.append(tokens.get(i));
                    
                    int endIndex = tokenEndIndex.get(i).intValue();

                    for (int j = 1; j < k; ++j)
                    {
                        sb.append(' ');
                        sb.append(tokens.get(i + j));
                        
                        endIndex = tokenEndIndex.get(i + j).intValue();
                    }
                    
                    String token = sb.toString();
                    
//                    System.out.println("=> " + token);
                    
                    int tindex = tokenIndex.get(i).intValue();
                    
                    distinctTokens.add(token);
                }
            }        
        }        
        
        return distinctTokens;
    }
    
    public Set<String> getDistinctTokens(String textToTokenize, char[] delimch, int maxTokensInGroup)
    {
        List<String> tokens = new ArrayList<String>(); 
        List<Integer> tokenIndex = new ArrayList<Integer>();
        List<Integer> tokenEndIndex = new ArrayList<Integer>();
        Set<String> distinctTokens = new LinkedHashSet<String>();
        
        
        // custom tokenization routine
//        char[] delimch = delim.toCharArray();
        char[] inputch = textToTokenize.toCharArray();
        
        int startIndex = -1;
        
        for (int i = 0; i < inputch.length; ++i)
        {
            if (CommonUtil.arrayContainsValue(delimch, inputch[i]))
            {
                if (startIndex == -1)
                {
                    continue;
                }
                else
                {
                    tokenIndex.add(new Integer(startIndex));
                    tokenEndIndex.add(new Integer(i));
                    tokens.add(textToTokenize.substring(startIndex, i));
                    startIndex = -1;
                }                
            }
            else
            {
                if (startIndex == -1)
                {
                    startIndex = i;
                }
            }
        }
        
        if (startIndex != -1) // get the last token
        {
            tokenIndex.add(new Integer(startIndex));
            tokenEndIndex.add(new Integer(inputch.length));
            tokens.add(textToTokenize.substring(startIndex, inputch.length));
        }
        
        // end of custom tokenization routine
        
        
        // Collect groups of 1, 2, 3, 4, 5, and 6 tokens.
        int n = maxTokensInGroup;
        
        for (int k = n; k >= 1; --k)
        {
            
            for (int i = 0; i < tokens.size(); ++i)
            {
                if (i < (tokens.size() - (k - 1)))
                {
                    StringBuilder sb = new StringBuilder();
                    sb.append(tokens.get(i));
                    
                    int endIndex = tokenEndIndex.get(i).intValue();

                    for (int j = 1; j < k; ++j)
                    {
                        sb.append(' ');
                        sb.append(tokens.get(i + j));
                        
                        endIndex = tokenEndIndex.get(i + j).intValue();
                    }
                    
                    String token = sb.toString();
                    
                    int tindex = tokenIndex.get(i).intValue();
                    
                    distinctTokens.add(token);
                }
            }        
        }        
        
        return distinctTokens;
    }    
    
    public Set<String> getDistinctTokens(String textToTokenize, Set<Character> delimchset, int maxTokensInGroup)
    {
        List<String> tokens = new ArrayList<String>(); 
        List<Integer> tokenIndex = new ArrayList<Integer>();
        List<Integer> tokenEndIndex = new ArrayList<Integer>();
        Set<String> distinctTokens = new LinkedHashSet<String>();
        
        
        // custom tokenization routine
//        char[] delimch = delim.toCharArray();
        char[] inputch = textToTokenize.toCharArray();
        
        int startIndex = -1;
        
        for (int i = 0; i < inputch.length; ++i)
        {
            if (delimchset.contains(new Character(inputch[i])))
            {
                if (startIndex == -1)
                {
                    continue;
                }
                else
                {
                    tokenIndex.add(new Integer(startIndex));
                    tokenEndIndex.add(new Integer(i));
                    tokens.add(textToTokenize.substring(startIndex, i));
                    startIndex = -1;
                }                
            }
            else
            {
                if (startIndex == -1)
                {
                    startIndex = i;
                }
            }
        }
        
        if (startIndex != -1) // get the last token
        {
            tokenIndex.add(new Integer(startIndex));
            tokenEndIndex.add(new Integer(inputch.length));
            tokens.add(textToTokenize.substring(startIndex, inputch.length));
        }
        
        // end of custom tokenization routine
        
        
        // Collect groups of 1, 2, 3, 4, 5, and 6 tokens.
        int n = maxTokensInGroup;
        
        for (int k = n; k >= 1; --k)
        {
            
            for (int i = 0; i < tokens.size(); ++i)
            {
                if (i < (tokens.size() - (k - 1)))
                {
                    StringBuilder sb = new StringBuilder();
                    sb.append(tokens.get(i));
                    
                    int endIndex = tokenEndIndex.get(i).intValue();

                    for (int j = 1; j < k; ++j)
                    {
                        sb.append(' ');
                        sb.append(tokens.get(i + j));
                        
                        endIndex = tokenEndIndex.get(i + j).intValue();
                    }
                    
                    String token = sb.toString();
                    
//                    System.out.println("=> " + token);
                    
                    int tindex = tokenIndex.get(i).intValue();
                    
                    distinctTokens.add(token);
                }
            }        
        }        
        
        return distinctTokens;
    }
    
    public static void main(String[] main)
    {
        
        String input = "U, MDS Test";
        Tokenizer tokenizer = new Tokenizer();
        List<TokenGroup> tgList = tokenizer.getAllTokens(input, TOKEN_DELIM_1, 6);
        for (int i = 0; i < tgList.size(); ++i)
        {
            TokenGroup tg = tgList.get(i);
            System.out.println("{" + tg.getToken() + "}, index: " + tg.getIndex() + ", size: " + tg.getSize());
        }
    }

}
