/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package net.infantai.util.string.similarity;

import java.util.HashSet;
import java.util.Set;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class SimilarityCheckerCosineByTokens implements SimilarityChecker
{
    
    private static final String delim = "\\W+";
    
    public static void main(String[] args)
    {
//        String str1 = "The quick brown fox, jumps over the lazy dog";
//        String str2 = "The quick brown fox, jumps over the lazy dog ff";
        String str1 = "Other autoimmune hemolytic anemias or something else not classified elsewhere";
        String str2 = "Autoimmune anemias";
        SimilarityCheckerCosineByTokens scc = new SimilarityCheckerCosineByTokens();
        System.out.println(scc.getStringSimilarity(str1.toLowerCase(), str2.toLowerCase()));
//        SimilarityCheckerCosine sccOld = new SimilarityCheckerCosine();
//        System.out.println(sccOld.getStringSimilarity(str1, str2));
    }
    
    private Set<String> getTokens(String input)
    {
        Set<String> tokens = new HashSet<String>();
        String[] tokenArr = input.split(delim);
        for (int i = 0; i < tokenArr.length; ++i)
        {
            tokens.add(tokenArr[i]);
        }
        return tokens;
    }

    @Override
    public double getStringSimilarity(String str1, String str2)
    {
        double cosineSimilarity = 0.0;
        if (str1.trim().length() == 0 || str2.trim().length() == 0)
        {
            return cosineSimilarity;
        }
        Set<String> str1Tokens = getTokens(str1);
        Set<String> str2Tokens = getTokens(str2);
        Set<String> allTokens = new HashSet<String>(str1Tokens.size() + str2Tokens.size());
        allTokens.addAll(str1Tokens);
        allTokens.addAll(str2Tokens);
        int commonTermCount = (str1Tokens.size() + str2Tokens.size()) - allTokens.size();
        cosineSimilarity = (double) commonTermCount / (Math.pow(str1Tokens.size(), 0.5f) * Math.pow(str2Tokens.size(), 0.5f));
        return Math.min(1.0f, cosineSimilarity);
    }
    
}
